//----------------------------------------------------------------- 2013-10-24 ------------
// Microchip SST25VF032B 32Mbitフラッシュメモリーテストプログラム
//
//  SSTMem   arduino pin No
// --------  --------------
// 1 CS       10
// 2 SO       12 MISO
// 3 WP       3.3V
// 4 GND      GND
// 5 SI       11 MOSI
// 6 SCK      13 SCK
// 7 HOLD     3.3V
// 8 VCC      3.3V
//-------------------------------------------------------- written by iizuka --------------
#include <SPI.h>

// EEPROM関連
#define SST_CS     10    // SSTメモリーのCSピン番号
#define SST_SO     12    // SSTメモリーのCSピン番号

//--------------- 型変換の共用体定義 -----------------------------
typedef union {
  unsigned long long_val;
  struct {
    byte LL;  // Lower address
    byte ML;
    byte MH;
    byte HH;  // Higher address
  } byte_val;    
} type_convert_long_byte; 

//------------------------------------------------------------------------------------------
// SST Memory write enable
//------------------------------------------------------------------------------------------
void WREN( void )
{
  digitalWrite( SST_CS , LOW  );
  SPI.transfer( 0x06 );
  digitalWrite( SST_CS , HIGH );
}

//------------------------------------------------------------------------------------------
// SST Memory write disable
//------------------------------------------------------------------------------------------
void WRDI( void )
{
  digitalWrite( SST_CS , LOW  );
  SPI.transfer( 0x04 );
  digitalWrite( SST_CS , HIGH );
}

//------------------------------------------------------------------------------------------
// Chip-Erase  メモリーの内容は 0xFF に初期化される
//------------------------------------------------------------------------------------------
void chip_erase( void )
{
  Serial.println("Chip-Erase start");
  WREN();
  digitalWrite( SST_CS , LOW  );
  SPI.transfer( 0x60 );
  digitalWrite( SST_CS , HIGH );
  WRDI();
  while( device_busy() ) {}
  Serial.println("Chip-Erase end");
}

//------------------------------------------------------------------------------------------
// 4kbyte(=0x1000) Sector-Erase
//------------------------------------------------------------------------------------------
void sector_erase( unsigned long addr )
{
  type_convert_long_byte sst_addr;
  sst_addr.long_val = addr;
  
  Serial.println("Sector-Erase start");  
  WREN();
  digitalWrite( SST_CS , LOW  );
  SPI.transfer( 0x20 );
  SPI.transfer( sst_addr.byte_val.MH );
  SPI.transfer( sst_addr.byte_val.ML );
  SPI.transfer( sst_addr.byte_val.LL ); 
  digitalWrite( SST_CS , HIGH );
  WRDI();
  while( device_busy() ) {}
  Serial.println("Sector-Erase end");  
}

//------------------------------------------------------------------------------------------
// 32kbyte(=0x8000) Block-Erase
//------------------------------------------------------------------------------------------
void blockr_erase_32k( unsigned long addr )
{
  type_convert_long_byte sst_addr;
  sst_addr.long_val = addr;

  Serial.println("Block-Erase 32k start");      
  WREN();
  digitalWrite( SST_CS , LOW  );
  SPI.transfer( 0x52 );
  SPI.transfer( sst_addr.byte_val.MH );
  SPI.transfer( sst_addr.byte_val.ML );
  SPI.transfer( sst_addr.byte_val.LL ); 
  digitalWrite( SST_CS , HIGH );
  WRDI();
  while( device_busy() ) {}
  Serial.println("Block-Erase 32k end"); 
}

//------------------------------------------------------------------------------------------
// 64kbyte(=0x10000) Block-Erase
//------------------------------------------------------------------------------------------
void blockr_erase_64k( unsigned long addr )
{
  type_convert_long_byte sst_addr;
  sst_addr.long_val = addr;
  
  Serial.println("Block-Erase 64k start");   
  WREN();
  digitalWrite( SST_CS , LOW  );
  SPI.transfer( 0xD8 );
  SPI.transfer( sst_addr.byte_val.MH );
  SPI.transfer( sst_addr.byte_val.ML );
  SPI.transfer( sst_addr.byte_val.LL ); 
  digitalWrite( SST_CS , HIGH );
  WRDI();
  while( device_busy() ) {}
  Serial.println("Block-Erase 64k end"); 
}

//------------------------------------------------------------------------------------------
// SST Memoryから読み出す　
//------------------------------------------------------------------------------------------
void sst_read( unsigned long addr, byte data[], byte len )
{
  byte index;
  type_convert_long_byte sst_addr;

  digitalWrite( SST_CS , LOW  );  
  SPI.transfer( 0x03 );           // 25MHz Read
  sst_addr.long_val = addr;

  SPI.transfer( sst_addr.byte_val.MH );
  SPI.transfer( sst_addr.byte_val.ML );
  SPI.transfer( sst_addr.byte_val.LL ); 
  for( index = 0 ; index < len ; index++ ) {
    data[index] = SPI.transfer(0);
  }
  digitalWrite( SST_CS , HIGH );
}

//------------------------------------------------------------------------------------------
// SST Memoryから１バイト読み出す　
//------------------------------------------------------------------------------------------
byte sst_read_byte( unsigned long addr )
{
  type_convert_long_byte sst_addr;
  byte data;

  digitalWrite( SST_CS , LOW  );  
  SPI.transfer( 0x03 );           // 25MHz Read
  sst_addr.long_val = addr;

  SPI.transfer( sst_addr.byte_val.MH );
  SPI.transfer( sst_addr.byte_val.ML );
  SPI.transfer( sst_addr.byte_val.LL ); 
  data = SPI.transfer(0);
  digitalWrite( SST_CS , HIGH );

  return data;
}

//------------------------------------------------------------------------------------------
// SST Memoryに１バイト書き込む　
//------------------------------------------------------------------------------------------
void sst_write_byte( unsigned long addr, byte data )
{
  unsigned long index;
  type_convert_long_byte sst_addr;
  
  WREN();
  digitalWrite( SST_CS , LOW  );   
  SPI.transfer( 0x02 );           
  sst_addr.long_val = addr;
  SPI.transfer( sst_addr.byte_val.MH );
  SPI.transfer( sst_addr.byte_val.ML );
  SPI.transfer( sst_addr.byte_val.LL ); 
  SPI.transfer( data );
  digitalWrite( SST_CS , HIGH );
  WRDI();
  while( device_busy() ) {}
}

//------------------------------------------------------------------------------------------
// SST Memoryに書き込む
//    arduinoでは実効速度が遅く通信速度がネックにならないためこの方法は速くならない
//------------------------------------------------------------------------------------------
void sst_write_word_HW( unsigned long addr, byte data[], byte len )
{
  byte index;
  byte index_ad;
  type_convert_long_byte sst_addr;

  EBSY();
  WREN();
  digitalWrite( SST_CS , LOW  );  
  SPI.transfer( 0xAD );            
  sst_addr.long_val = addr;
  SPI.transfer( sst_addr.byte_val.MH );
  SPI.transfer( sst_addr.byte_val.ML );
  SPI.transfer( sst_addr.byte_val.LL );
  index_ad = 0;
  for( index = 0 ; index < len ; index++ ) {
    if( index_ad == 2 ) {
      digitalWrite( SST_CS , HIGH );
      digitalWrite( SST_CS , LOW  );
      while( digitalRead( SST_SO ) == LOW ) {}
      digitalWrite( SST_CS , HIGH );
      digitalWrite( SST_CS , LOW  );
      SPI.transfer( 0xAD ); 
      index_ad = 0;
    }
    SPI.transfer( data[index] );
    index_ad++;
  }
  digitalWrite( SST_CS , HIGH );
  digitalWrite( SST_CS , LOW  );
  while( digitalRead( SST_SO ) == LOW ) {}
  digitalWrite( SST_CS , HIGH );
  digitalWrite( SST_CS , LOW  );
  WRDI();
  DBSY();
  while( device_busy() ) {}
}

//------------------------------------------------------------------------------------------
// SST Memoryに書き込む　
//------------------------------------------------------------------------------------------
void sst_write_word_SW( unsigned long addr, byte data[], byte len )
{
  byte index;
  byte index_ad;
  type_convert_long_byte sst_addr;

  WREN();
  digitalWrite( SST_CS , LOW  ); 
  SPI.transfer( 0xAD ); 
  sst_addr.long_val = addr;
  SPI.transfer( sst_addr.byte_val.MH );
  SPI.transfer( sst_addr.byte_val.ML );
  SPI.transfer( sst_addr.byte_val.LL );
  index_ad = 0;
  for( index = 0 ; index < len ; index++ ) {
    if( index_ad == 2 ) {
      digitalWrite( SST_CS , HIGH );
      while( device_busy() ) {}
      digitalWrite( SST_CS , LOW  );
      SPI.transfer( 0xAD ); 
      index_ad = 0;
    }
    SPI.transfer( data[index] );
    index_ad++;
  }
  digitalWrite( SST_CS , HIGH );
  while( device_busy() ) {}
  WRDI();
  while( device_busy() ) {}
}

//------------------------------------------------------------------------------------------
// ＡＡＩ書き込みのためSOピンによるステータス出力を可能にする
//------------------------------------------------------------------------------------------
void EBSY(void)
{
  digitalWrite( SST_CS , LOW  );   
  SPI.transfer( 0x70 );            
  digitalWrite( SST_CS , HIGH );
}

//------------------------------------------------------------------------------------------
// ＡＡＩ書き込みのためSOピンによるステータス不可にする
//------------------------------------------------------------------------------------------
void DBSY(void)
{
  digitalWrite( SST_CS , LOW  );   
  SPI.transfer( 0x80 );            
  digitalWrite( SST_CS , HIGH );
}

//------------------------------------------------------------------------------------------
// status registerを読み出す
//------------------------------------------------------------------------------------------
byte RDSR(void)
{
  byte data;
  
  digitalWrite( SST_CS , LOW  ); 
  SPI.transfer( 0x05 );         
  data = SPI.transfer( 0x00 );
  digitalWrite( SST_CS , HIGH );

  return data;
}

//------------------------------------------------------------------------------------------
// device busy検査
//------------------------------------------------------------------------------------------
boolean device_busy(void)
{
  if( (RDSR() & 0x01) == 0x01 ) {
//    Serial.println("BUSY");
    return true;
  } else {
    return false;
  }
}

//------------------------------------------------------------------------------------------
// status registerに書き込む
//------------------------------------------------------------------------------------------
void WRSR(byte reg_data)
{
  WREN();
  digitalWrite( SST_CS , LOW  ); 
  SPI.transfer( 0x01 );       
  SPI.transfer( reg_data );
  digitalWrite( SST_CS , HIGH );
}

//------------------------------------------------------------------------------------------
// SST Memoryの WriteProtect を解除する
//------------------------------------------------------------------------------------------
void device_protect_disable()
{
  WRSR( 0b00000000 );
}

//------------------------------------------------------------------------------------------
// SST Memoryの WriteProtect を設定する
//------------------------------------------------------------------------------------------
void device_protect_enable()
{
  WRSR( 0b10111100 );
}

//------------------------------------------------------------------------------------------
// arduino setup
//------------------------------------------------------------------------------------------  
void setup()
{
  pinMode(SST_CS, OUTPUT);
  pinMode(SST_SO, INPUT );
  digitalWrite( SST_CS , HIGH );   // disable SST Memory
  
  delay(1000);
  
  SPI.begin();
//  SPI.setClockDivider( SPI_CLOCK_DIV128 );  // オシロスコープで波形を見るときはクロックを遅くすると良い
//  SPI.setBitOrder( MSBFIRST );              // デフォルトでMSBFIRSTなので無くてもあっても良い
//  SPI.setDataMode( SPI_MODE0 );             // これもデフォルト設定なのであっても無くても良い
  
  Serial.begin(57600);

  Serial.println("--------- TEST START ------------");
  device_protect_disable(); 
  chip_erase();
}

//------------------------------------------------------------------------------------------
// arduino loop
//------------------------------------------------------------------------------------------  
void loop()
{
  byte data_w;
  byte data_r;
  unsigned long addr;
  byte data_array_w[16];
  byte data_array_r[16];
  unsigned long tm;

  randomSeed(analogRead(0));

  //------ １バイト単位の読み書き ------------------------------------------
  Serial.println("--------- TEST 1 ---------------");  
  Serial.println("byte write/read start");
  tm = millis();
  for( addr = 0x0 ; addr < 0x20000 ; addr++ ) {  // 約３０秒かかる
    data_w = random(0,255);  // この処理は遅い
    sst_write_byte( addr, data_w );
    data_r = sst_read_byte(addr);
    if( data_w != data_r ) {
      Serial.print("error addr=");  Serial.print(addr,HEX);
      Serial.print(" write data="); Serial.print(data_w,HEX);
      Serial.print(" read data=" ); Serial.print(data_r,HEX); Serial.println("");
    }  
  }
  Serial.print("byte write/read end");
  Serial.print(" Time:"); Serial.print((millis() - tm )/1000); Serial.println("sec"); 

  //------ ＡＡＩモード書き込み ソフトウェアによるBUSY判定 -------------------
  Serial.println("--------- TEST 2 ---------------");  
  Serial.println("S/W array write/read start"); 
  for( byte i = 0 ; i < 16 ; i++ ) {
    data_array_w[i] = random(0,255);
  } 
  tm = millis();
  for( addr = 0x20000 ; addr < 0x40000 ; addr += 16 ) {  
    sst_write_word_HW( addr, data_array_w, 16 );
    sst_read( addr, data_array_r, 16 );
    for( byte i = 0 ; i < 16 ; i++ ) {
      if( data_array_w[i] != data_array_r[i] ) {
        Serial.print("error addr=");  Serial.print(addr,HEX);
        Serial.print(" index=");      Serial.print(i);
        Serial.print(" write data="); Serial.print(data_array_w[i],HEX);
        Serial.print(" read data=" ); Serial.print(data_array_r[i],HEX); Serial.println("");
      }
    }
  }
  Serial.print("S/W array write/read end"); 
  Serial.print(" Time:");  Serial.print((millis() - tm )/1000); Serial.println("sec");

  //------ ＡＡＩモード書き込み ハードウェアによるBUSY判定 -------------------
  Serial.println("--------- TEST 3 ---------------");    
  Serial.println("H/W array write/read start"); 
  for( byte i = 0 ; i < 16 ; i++ ) {
    data_array_w[i] = random(0,255);
  } 
  tm = millis();
  for( addr = 0x40000 ; addr < 0x60000 ; addr += 16 ) {  
    sst_write_word_SW( addr, data_array_w, 16 );
    sst_read( addr, data_array_r, 16 );
    for( byte i = 0 ; i < 16 ; i++ ) {
      if( data_array_w[i] != data_array_r[i] ) {
        Serial.print("error addr=");  Serial.print(addr,HEX);
        Serial.print(" index=");      Serial.print(i);
        Serial.print(" write data="); Serial.print(data_array_w[i],HEX);
        Serial.print(" read data=" ); Serial.print(data_array_r[i],HEX); Serial.println("");
      }
    }
  }
  Serial.print("H/W array write/read end"); 
  Serial.print(" Time:");  Serial.print((millis() - tm )/1000); Serial.println("sec");

  //------ １バイト単位の読み書き ------------------------------------------
  // ERASEしないで上書きするとエラーになる
  Serial.println("--------- TEST 4 ---------------");  
  Serial.println("byte write/read start");
  tm = millis();
  for( addr = 0x0 ; addr < 0x5 ; addr++ ) {  
    data_w = random(0,255);
    sst_write_byte( addr, data_w );
    data_r = sst_read_byte(addr);
    if( data_w != data_r ) {
      Serial.print("error addr=");  Serial.print(addr,HEX);
      Serial.print(" write data="); Serial.print(data_w,HEX);
      Serial.print(" read data=" ); Serial.print(data_r,HEX); Serial.println("");
    }  
  }
  Serial.print("byte write/read end");
  Serial.print(" Time:"); Serial.print((millis() - tm )/1000); Serial.println("sec");

  //------ １バイト単位の読み書き ------------------------------------------
  // ERASEしてから書き込む必要がある
  Serial.println("--------- TEST 5 ---------------");  
  sector_erase( 0x0 );
  Serial.println("byte write/read start");
  tm = millis();
  for( addr = 0x0 ; addr < 0x5 ; addr++ ) {  
    data_w = random(0,255);
    sst_write_byte( addr, data_w );
    data_r = sst_read_byte(addr);
    if( data_w != data_r ) {
      Serial.print("error addr=");  Serial.print(addr,HEX);
      Serial.print(" write data="); Serial.print(data_w,HEX);
      Serial.print(" read data=" ); Serial.print(data_r,HEX); Serial.println("");
    }  
  }
  Serial.print("byte write/read end");
  Serial.print(" Time:"); Serial.print((millis() - tm )/1000); Serial.println("sec");  

  //------ １バイト単位の読み書き ------------------------------------------
  // 強引に上書きするとどうなるかテスト
  // 消去しないで上書きするとＡＮＤが取られる
  // 消去は全ビットセット　書き込みによりビットクリヤは出来るがビットセットが出来ない
  Serial.println("--------- TEST 6 ---------------");  
  sector_erase( 0x0 );
  Serial.println("write BIN 11111100");
  sst_write_byte( 0x0, 0b11111100 );
  Serial.println("write BIN 11001111");
  sst_write_byte( 0x0, 0b11001111 );
  data_r = sst_read_byte( 0x0 );
  Serial.print("read data=" ); Serial.print(data_r,BIN); Serial.println("");
  
  Serial.println("--------- TEST END -------------");
  device_protect_enable();
  while(true) {}
}

//------- E N D ------------------------------------------------------------------------

